#    Copyright 2015 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import timeutils
import pytz
import six

from cinder.db.sqlalchemy import models
from cinder import exception
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import objects as test_objects
from cinder.tests.unit import utils


fake_backup = {
    'id': fake.BACKUP_ID,
    'volume_id': fake.VOLUME_ID,
    'status': fields.BackupStatus.CREATING,
    'size': 1,
    'display_name': 'fake_name',
    'display_description': 'fake_description',
    'user_id': fake.USER_ID,
    'project_id': fake.PROJECT_ID,
    'temp_volume_id': None,
    'temp_snapshot_id': None,
    'snapshot_id': None,
    'data_timestamp': None,
    'restore_volume_id': None,
    'backup_metadata': {},
}

vol_props = {'status': 'available', 'size': 1}
fake_vol = fake_volume.fake_db_volume(**vol_props)
snap_props = {'status': fields.BackupStatus.AVAILABLE,
              'volume_id': fake_vol['id'],
              'expected_attrs': ['metadata']}
fake_snap = fake_snapshot.fake_db_snapshot(**snap_props)


class TestBackup(test_objects.BaseObjectsTestCase):

    @mock.patch('cinder.db.get_by_id', return_value=fake_backup)
    def test_get_by_id(self, backup_get):
        backup = objects.Backup.get_by_id(self.context, fake.USER_ID)
        self._compare(self, fake_backup, backup)
        backup_get.assert_called_once_with(self.context, models.Backup,
                                           fake.USER_ID)

    @mock.patch('cinder.db.sqlalchemy.api.model_query')
    def test_get_by_id_no_existing_id(self, model_query):
        query = mock.Mock()
        filter_by = mock.Mock()
        query_options = mock.Mock()
        filter_by.first.return_value = None
        query_options.filter_by.return_value = filter_by
        query.options.return_value = query_options
        model_query.return_value = query
        self.assertRaises(exception.BackupNotFound, objects.Backup.get_by_id,
                          self.context, 123)

    @mock.patch('cinder.db.backup_create', return_value=fake_backup)
    def test_create(self, backup_create):
        backup = objects.Backup(context=self.context)
        backup.create()
        self.assertEqual(fake_backup['id'], backup.id)
        self.assertEqual(fake_backup['volume_id'], backup.volume_id)

    @mock.patch('cinder.db.backup_update')
    def test_save(self, backup_update):
        backup = objects.Backup._from_db_object(
            self.context, objects.Backup(), fake_backup)
        backup.display_name = 'foobar'
        backup.save()
        backup_update.assert_called_once_with(self.context, backup.id,
                                              {'display_name': 'foobar'})

    @mock.patch('cinder.db.backup_metadata_update',
                return_value={'key1': 'value1'})
    @mock.patch('cinder.db.backup_update')
    def test_save_with_metadata(self, backup_update, metadata_update):
        backup = objects.Backup._from_db_object(
            self.context, objects.Backup(), fake_backup)

        backup.metadata = {'key1': 'value1'}
        self.assertEqual({'metadata': {'key1': 'value1'}},
                         backup.obj_get_changes())
        backup.save()
        metadata_update.assert_called_once_with(self.context, backup.id,
                                                {'key1': 'value1'}, True)

    @mock.patch('oslo_utils.timeutils.utcnow', return_value=timeutils.utcnow())
    @mock.patch('cinder.db.sqlalchemy.api.backup_destroy')
    def test_destroy(self, backup_destroy, utcnow_mock):
        backup_destroy.return_value = {
            'status': fields.BackupStatus.DELETED,
            'deleted': True,
            'deleted_at': utcnow_mock.return_value}
        backup = objects.Backup(context=self.context, id=fake.BACKUP_ID)
        backup.destroy()
        self.assertTrue(backup_destroy.called)
        admin_context = backup_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)
        self.assertTrue(backup.deleted)
        self.assertEqual(fields.BackupStatus.DELETED, backup.status)
        self.assertEqual(utcnow_mock.return_value.replace(tzinfo=pytz.UTC),
                         backup.deleted_at)

    def test_obj_field_temp_volume_snapshot_id(self):
        backup = objects.Backup(context=self.context,
                                temp_volume_id='2',
                                temp_snapshot_id='3')
        self.assertEqual('2', backup.temp_volume_id)
        self.assertEqual('3', backup.temp_snapshot_id)

    def test_obj_field_snapshot_id(self):
        backup = objects.Backup(context=self.context,
                                snapshot_id='2')
        self.assertEqual('2', backup.snapshot_id)

    def test_obj_field_restore_volume_id(self):
        backup = objects.Backup(context=self.context,
                                restore_volume_id='2')
        self.assertEqual('2', backup.restore_volume_id)

    def test_obj_field_metadata(self):
        backup = objects.Backup(context=self.context,
                                metadata={'test_key': 'test_value'})
        self.assertEqual({'test_key': 'test_value'}, backup.metadata)

    @mock.patch('cinder.objects.backup.Backup.get_by_id',
                return_value=None)
    def test_obj_field_parent(self, mock_lzy_ld):
        backup = objects.Backup(context=self.context,
                                parent_id=None)
        self.assertIsNone(backup.parent)

        # Bug #1862635: should trigger a lazy load
        backup = objects.Backup(context=self.context,
                                parent_id=fake.UUID5)
        _ = backup.parent
        mock_lzy_ld.assert_called_once()

    def test_import_record(self):
        utils.replace_obj_loader(self, objects.Backup)
        backup = objects.Backup(context=self.context, id=fake.BACKUP_ID,
                                parent_id=None,
                                num_dependent_backups=0)
        export_string = backup.encode_record()
        imported_backup = objects.Backup.decode_record(export_string)

        # Make sure we don't lose data when converting from string
        self.assertDictEqual(self._expected_backup(backup), imported_backup)

    @mock.patch('cinder.db.get_by_id', return_value=fake_backup)
    def test_import_record_w_parent(self, backup_get):
        full_backup = objects.Backup.get_by_id(self.context, fake.USER_ID)
        self._compare(self, fake_backup, full_backup)

        utils.replace_obj_loader(self, objects.Backup)
        incr_backup = objects.Backup(context=self.context,
                                     id=fake.BACKUP2_ID,
                                     parent=full_backup,
                                     parent_id=full_backup['id'],
                                     num_dependent_backups=0)
        export_string = incr_backup.encode_record()
        imported_backup = objects.Backup.decode_record(export_string)

        # Make sure we don't lose data when converting from string
        self.assertDictEqual(self._expected_backup(incr_backup),
                             imported_backup)

    def test_import_record_additional_info(self):
        utils.replace_obj_loader(self, objects.Backup)
        backup = objects.Backup(context=self.context, id=fake.BACKUP_ID,
                                parent_id=None,
                                num_dependent_backups=0)
        extra_info = {'driver': {'key1': 'value1', 'key2': 'value2'}}
        extra_info_copy = extra_info.copy()
        export_string = backup.encode_record(extra_info=extra_info)
        imported_backup = objects.Backup.decode_record(export_string)

        # Dictionary passed should not be modified
        self.assertDictEqual(extra_info_copy, extra_info)

        # Make sure we don't lose data when converting from string and that
        # extra info is still there
        expected = self._expected_backup(backup)
        expected['extra_info'] = extra_info
        self.assertDictEqual(expected, imported_backup)

    def _expected_backup(self, backup):
        record = {name: field.to_primitive(backup, name, getattr(backup, name))
                  for name, field in backup.fields.items() if name != 'parent'}
        return record

    def test_import_record_additional_info_cant_overwrite(self):
        utils.replace_obj_loader(self, objects.Backup)
        backup = objects.Backup(context=self.context, id=fake.BACKUP_ID,
                                parent_id=None,
                                num_dependent_backups=0)
        export_string = backup.encode_record(id='fake_id')
        imported_backup = objects.Backup.decode_record(export_string)

        # Make sure the extra_info can't overwrite basic data
        self.assertDictEqual(self._expected_backup(backup), imported_backup)

    def test_import_record_decoding_error(self):
        export_string = '123456'
        self.assertRaises(exception.InvalidInput,
                          objects.Backup.decode_record,
                          export_string)

    def test_import_record_parsing_error(self):
        export_string = ''
        self.assertRaises(exception.InvalidInput,
                          objects.Backup.decode_record,
                          export_string)

    @mock.patch('cinder.db.sqlalchemy.api.backup_get')
    def test_refresh(self, backup_get):
        db_backup1 = fake_backup.copy()
        db_backup2 = db_backup1.copy()
        db_backup2['display_name'] = 'foobar'

        # On the second backup_get, return the backup with an updated
        # display_name
        backup_get.side_effect = [db_backup1, db_backup2]
        backup = objects.Backup.get_by_id(self.context, fake.BACKUP_ID)
        self._compare(self, db_backup1, backup)

        # display_name was updated, so a backup refresh should have a new value
        # for that field
        backup.refresh()
        self._compare(self, db_backup2, backup)
        if six.PY3:
            call_bool = mock.call.__bool__()
        else:
            call_bool = mock.call.__nonzero__()
        backup_get.assert_has_calls([mock.call(self.context, fake.BACKUP_ID),
                                     call_bool,
                                     mock.call(self.context, fake.BACKUP_ID)])


class TestBackupList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.backup_get_all', return_value=[fake_backup])
    def test_get_all(self, backup_get_all):
        backups = objects.BackupList.get_all(self.context)
        self.assertEqual(1, len(backups))
        TestBackup._compare(self, fake_backup, backups[0])

    @mock.patch('cinder.db.backup_get_all_by_project',
                return_value=[fake_backup])
    def test_get_all_by_project(self, get_all_by_project):
        backups = objects.BackupList.get_all_by_project(
            self.context, self.project_id)
        self.assertEqual(1, len(backups))
        TestBackup._compare(self, fake_backup, backups[0])

    @mock.patch('cinder.db.backup_get_all_by_host',
                return_value=[fake_backup])
    def test_get_all_by_host(self, get_all_by_host):
        backups = objects.BackupList.get_all_by_host(self.context, "fake_host")
        self.assertEqual(1, len(backups))
        TestBackup._compare(self, fake_backup, backups[0])

    @mock.patch('cinder.db.backup_get_all', return_value=[fake_backup])
    def test_get_all_tenants(self, backup_get_all):
        search_opts = {'all_tenants': 1}
        backups = objects.BackupList.get_all(self.context, search_opts)
        self.assertEqual(1, len(backups))
        TestBackup._compare(self, fake_backup, backups[0])

    @mock.patch('cinder.db.backup_get_all_by_volume',
                return_value=[fake_backup])
    def test_get_all_by_volume(self, get_all_by_volume):
        backups = objects.BackupList.get_all_by_volume(
            self.context, fake.VOLUME_ID, 'fake_proj')
        self.assertEqual(1, len(backups))
        get_all_by_volume.assert_called_once_with(self.context,
                                                  fake.VOLUME_ID,
                                                  'fake_proj',
                                                  None)
        TestBackup._compare(self, fake_backup, backups[0])


class BackupDeviceInfoTestCase(test_objects.BaseObjectsTestCase):
    def setUp(self):
        super(BackupDeviceInfoTestCase, self).setUp()
        self.vol_obj = fake_volume.fake_volume_obj(self.context, **vol_props)
        self.snap_obj = fake_snapshot.fake_snapshot_obj(self.context,
                                                        **snap_props)
        self.backup_device_dict = {'secure_enabled': False,
                                   'is_snapshot': False, }

    @mock.patch('cinder.db.volume_get', return_value=fake_vol)
    def test_from_primitive_with_volume(self, mock_fake_vol):
        vol_obj = self.vol_obj
        self.backup_device_dict['backup_device'] = vol_obj
        backup_device_info = objects.BackupDeviceInfo.from_primitive(
            self.backup_device_dict, self.context)
        self.assertFalse(backup_device_info.is_snapshot)
        self.assertEqual(self.backup_device_dict['secure_enabled'],
                         backup_device_info.secure_enabled)
        self.assertEqual(vol_obj, backup_device_info.volume)

        self.backup_device_dict['backup_device'] = fake_vol
        backup_device_info = objects.BackupDeviceInfo.from_primitive(
            self.backup_device_dict, self.context)
        vol_obj_from_db = objects.Volume._from_db_object(self.context,
                                                         objects.Volume(),
                                                         fake_vol)
        self.assertEqual(vol_obj_from_db, backup_device_info.volume)

    @mock.patch('cinder.db.snapshot_get', return_value=fake_snap)
    def test_from_primitive_with_snapshot(self, mock_fake_snap):
        snap_obj = self.snap_obj
        self.backup_device_dict['is_snapshot'] = True
        self.backup_device_dict['backup_device'] = snap_obj
        backup_device_info = objects.BackupDeviceInfo.from_primitive(
            self.backup_device_dict, self.context, expected_attrs=['metadata'])
        self.assertTrue(backup_device_info.is_snapshot)
        self.assertEqual(self.backup_device_dict['secure_enabled'],
                         backup_device_info.secure_enabled)
        self.assertEqual(snap_obj, backup_device_info.snapshot)

        self.backup_device_dict['backup_device'] = fake_snap
        backup_device_info = objects.BackupDeviceInfo.from_primitive(
            self.backup_device_dict, self.context, expected_attrs=['metadata'])
        self.assertEqual(snap_obj, backup_device_info.snapshot)

    @mock.patch('cinder.db.volume_get', return_value=fake_vol)
    def test_to_primitive_with_volume(self, mock_fake_vol):
        vol_obj = self.vol_obj
        self.backup_device_dict['backup_device'] = fake_vol
        backup_device_info = objects.BackupDeviceInfo()
        backup_device_info.volume = vol_obj
        backup_device_info.secure_enabled = (
            self.backup_device_dict['secure_enabled'])

        backup_device_ret_dict = backup_device_info.to_primitive(self.context)
        self.assertEqual(self.backup_device_dict['secure_enabled'],
                         backup_device_ret_dict['secure_enabled'])
        self.assertFalse(backup_device_ret_dict['is_snapshot'])
        self.assertEqual(self.backup_device_dict['backup_device'],
                         backup_device_ret_dict['backup_device'])

    @mock.patch('cinder.db.snapshot_get', return_value=fake_snap)
    def test_to_primitive_with_snapshot(self, mock_fake_snap):
        snap_obj = self.snap_obj
        backup_device_info = objects.BackupDeviceInfo()
        backup_device_info.snapshot = snap_obj
        backup_device_info.secure_enabled = (
            self.backup_device_dict['secure_enabled'])

        backup_device_ret_dict = backup_device_info.to_primitive(self.context)
        self.assertEqual(self.backup_device_dict['secure_enabled'],
                         backup_device_ret_dict['secure_enabled'])
        self.assertTrue(backup_device_ret_dict['is_snapshot'])
        # NOTE(sborkows): since volume in sqlalchemy snapshot is a sqlalchemy
        # object too, to compare snapshots we need to convert their volumes to
        # dicts.
        snap_actual_dict = fake_snap
        snap_ref_dict = backup_device_ret_dict['backup_device']
        snap_actual_dict['volume'] = self.vol_obj.obj_to_primitive()
        snap_ref_dict['volume'] = snap_ref_dict['volume']
        self.assertEqual(snap_actual_dict, snap_ref_dict)

    def test_is_snapshot_both_volume_and_snapshot_raises_error(self):
        snap = self.snap_obj
        vol = self.vol_obj
        backup_device_info = objects.BackupDeviceInfo()
        backup_device_info.snapshot = snap
        backup_device_info.volume = vol
        backup_device_info.secure_enabled = (
            self.backup_device_dict['secure_enabled'])
        self.assertRaises(exception.ProgrammingError, getattr,
                          backup_device_info, 'is_snapshot')

    def test_is_snapshot_neither_volume_nor_snapshot_raises_error(self):
        backup_device_info = objects.BackupDeviceInfo()
        backup_device_info.secure_enabled = (
            self.backup_device_dict['secure_enabled'])
        self.assertRaises(exception.ProgrammingError, getattr,
                          backup_device_info, 'is_snapshot')

    def test_device_obj_with_volume(self):
        vol = self.vol_obj
        backup_device_info = objects.BackupDeviceInfo()
        backup_device_info.volume = vol
        backup_device_info.secure_enabled = (
            self.backup_device_dict['secure_enabled'])
        backup_device_obj = backup_device_info.device_obj
        self.assertIsInstance(backup_device_obj, objects.Volume)
        self.assertEqual(vol, backup_device_obj)

    def test_device_obj_with_snapshot(self):
        snap = self.snap_obj
        backup_device_info = objects.BackupDeviceInfo()
        backup_device_info.snapshot = snap
        backup_device_info.secure_enabled = (
            self.backup_device_dict['secure_enabled'])
        backup_device_obj = backup_device_info.device_obj
        self.assertIsInstance(backup_device_obj, objects.Snapshot)
        self.assertEqual(snap, backup_device_obj)



# Copyright (c) - 2014, Clinton Knight.  All rights reserved.
# Copyright (c) - 2015, Tom Barron.  All rights reserved.
# Copyright (c) - 2016 Chuck Fouts. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from lxml import etree

from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api


VOLUME_ID = 'f10d1a84-9b7b-427e-8fec-63c48b509a56'
LUN_ID = 'ee6b4cc7-477b-4016-aa0c-7127b4e3af86'
LUN_HANDLE = 'fake_lun_handle'
LUN_NAME = 'lun1'
LUN_SIZE = 3
LUN_TABLE = {LUN_NAME: None}
SIZE = 1024
HOST_NAME = 'fake.host.name'
BACKEND_NAME = 'fake_backend_name'
POOL_NAME = 'aggr1'
SHARE_IP = '192.168.99.24'
IPV6_ADDRESS = 'fe80::6e40:8ff:fe8a:130'
EXPORT_PATH = '/fake/export/path'
NFS_SHARE = '%s:%s' % (SHARE_IP, EXPORT_PATH)
NFS_SHARE_IPV6 = '[%s]:%s' % (IPV6_ADDRESS, EXPORT_PATH)
HOST_STRING = '%s@%s#%s' % (HOST_NAME, BACKEND_NAME, POOL_NAME)
NFS_HOST_STRING = '%s@%s#%s' % (HOST_NAME, BACKEND_NAME, NFS_SHARE)
AGGREGATE = 'aggr1'
FLEXVOL = 'openstack-flexvol'
NFS_FILE_PATH = 'nfsvol'
PATH = '/vol/%s/%s' % (POOL_NAME, LUN_NAME)
IMAGE_FILE_ID = 'img-cache-imgid'
PROVIDER_LOCATION = 'fake_provider_location'
NFS_HOST = 'nfs-host1'
NFS_SHARE_PATH = '/export'
NFS_EXPORT_1 = '%s:%s' % (NFS_HOST, NFS_SHARE_PATH)
NFS_EXPORT_2 = 'nfs-host2:/export'
MOUNT_POINT = '/mnt/nfs'
ATTACHED = 'attached'
DETACHED = 'detached'
LUN_METADATA = {
    'OsType': None,
    'SpaceReserved': 'true',
    'Path': PATH,
    'Qtree': None,
    'Volume': POOL_NAME,
}
VOLUME = {
    'name': LUN_NAME,
    'size': SIZE,
    'id': VOLUME_ID,
    'host': HOST_STRING,
    'attach_status': DETACHED,
}
NFS_VOLUME = {
    'name': NFS_FILE_PATH,
    'size': SIZE,
    'id': VOLUME_ID,
    'host': NFS_HOST_STRING,
    'provider_location': PROVIDER_LOCATION,
}

FAKE_MANAGE_VOLUME = {
    'name': 'volume-new-managed-123',
    'id': 'volume-new-managed-123',
}

FAKE_IMAGE_LOCATION = (
    None,
    [
        # valid metadata
        {
            'metadata': {
                'share_location': 'nfs://host/path',
                'mountpoint': '/opt/stack/data/glance',
                'id': 'abc-123',
                'type': 'nfs'
            },
            'url': 'file:///opt/stack/data/glance/image-id-0'
        },
        # missing metadata
        {
            'metadata': {},
            'url': 'file:///opt/stack/data/glance/image-id-1'
        },
        # missing location_type
        {
            'metadata': {'location_type': None},
            'url': 'file:///opt/stack/data/glance/image-id-2'
        },
        # non-nfs location_type
        {
            'metadata': {'location_type': 'not-NFS'},
            'url': 'file:///opt/stack/data/glance/image-id-3'
        },
        # missing share_location
        {
            'metadata': {'location_type': 'nfs', 'share_location': None},
            'url': 'file:///opt/stack/data/glance/image-id-4'},
        # missing mountpoint
        {
            'metadata': {
                'location_type': 'nfs',
                'share_location': 'nfs://host/path',
                # Pre-kilo we documented "mount_point"
                'mount_point': '/opt/stack/data/glance'
            },
            'url': 'file:///opt/stack/data/glance/image-id-5'
        },
        # Valid metadata
        {
            'metadata':
                {
                    'share_location': 'nfs://host/path',
                    'mountpoint': '/opt/stack/data/glance',
                    'id': 'abc-123',
                    'type': 'nfs',
                },
            'url': 'file:///opt/stack/data/glance/image-id-6'
        }
    ]
)

NETAPP_VOLUME = 'fake_netapp_volume'

VFILER = 'fake_netapp_vfiler'

UUID1 = '12345678-1234-5678-1234-567812345678'
LUN_PATH = '/vol/vol0/%s' % LUN_NAME

VSERVER_NAME = 'openstack-vserver'

FC_VOLUME = {'name': 'fake_volume'}

FC_INITIATORS = ['21000024ff406cc3', '21000024ff406cc2']
FC_FORMATTED_INITIATORS = ['21:00:00:24:ff:40:6c:c3',
                           '21:00:00:24:ff:40:6c:c2']

FC_TARGET_WWPNS = ['500a098280feeba5', '500a098290feeba5',
                   '500a098190feeba5', '500a098180feeba5']

FC_FORMATTED_TARGET_WWPNS = ['50:0a:09:82:80:fe:eb:a5',
                             '50:0a:09:82:90:fe:eb:a5',
                             '50:0a:09:81:90:fe:eb:a5',
                             '50:0a:09:81:80:fe:eb:a5']

FC_CONNECTOR = {'ip': '1.1.1.1',
                'host': 'fake_host',
                'wwnns': ['20000024ff406cc3', '20000024ff406cc2'],
                'wwpns': ['21000024ff406cc3', '21000024ff406cc2']}

FC_I_T_MAP = {'21000024ff406cc3': ['500a098280feeba5', '500a098290feeba5'],
              '21000024ff406cc2': ['500a098190feeba5', '500a098180feeba5']}

FC_I_T_MAP_COMPLETE = {'21000024ff406cc3': FC_TARGET_WWPNS,
                       '21000024ff406cc2': FC_TARGET_WWPNS}

FC_FABRIC_MAP = {'fabricB':
                 {'target_port_wwn_list':
                  ['500a098190feeba5', '500a098180feeba5'],
                  'initiator_port_wwn_list': ['21000024ff406cc2']},
                 'fabricA':
                 {'target_port_wwn_list':
                  ['500a098290feeba5', '500a098280feeba5'],
                  'initiator_port_wwn_list': ['21000024ff406cc3']}}

FC_TARGET_INFO = {'driver_volume_type': 'fibre_channel',
                  'data': {'target_lun': 1,
                           'initiator_target_map': FC_I_T_MAP,
                           'target_wwn': FC_TARGET_WWPNS,
                           'target_discovered': True}}

FC_TARGET_INFO_EMPTY = {'driver_volume_type': 'fibre_channel', 'data': {}}

FC_TARGET_INFO_UNMAP = {'driver_volume_type': 'fibre_channel',
                        'data': {'target_wwn': FC_TARGET_WWPNS,
                                 'initiator_target_map': FC_I_T_MAP}}

ISCSI_ONE_MAP_LIST = [{'initiator-group': 'openstack-faketgt1',
                       'vserver': 'vserver_123', 'lun-id': '1'}]
ISCSI_MULTI_MAP_LIST = [{'initiator-group': 'openstack-faketgt1',
                        'vserver': 'vserver_123', 'lun-id': '1'},
                        {'initiator-group': 'openstack-faketgt2',
                         'vserver': 'vserver_123', 'lun-id': '2'}
                        ]
ISCSI_EMPTY_MAP_LIST = []

IGROUP1_NAME = 'openstack-igroup1'

IGROUP1 = {
    'initiator-group-os-type': 'linux',
    'initiator-group-type': 'fcp',
    'initiator-group-name': IGROUP1_NAME,
}

ISCSI_VOLUME = {
    'name': 'fake_volume',
    'id': 'fake_id',
    'provider_auth': 'fake provider auth',
}

ISCSI_LUN = {'name': ISCSI_VOLUME, 'lun_id': 42}

ISCSI_SERVICE_IQN = 'fake_iscsi_service_iqn'

ISCSI_CONNECTION_PROPERTIES = {
    'data': {
        'auth_method': 'fake_method',
        'auth_password': 'auth',
        'auth_username': 'provider',
        'discovery_auth_method': 'fake_method',
        'discovery_auth_username': 'provider',
        'discovery_auth_password': 'auth',
        'target_discovered': False,
        'target_iqn': ISCSI_SERVICE_IQN,
        'target_lun': 42,
        'target_portal': '1.2.3.4:3260',
        'volume_id': 'fake_id',
    },
    'driver_volume_type': 'iscsi',
}

ISCSI_CONNECTOR = {
    'ip': '1.1.1.1',
    'host': 'fake_host',
    'initiator': 'fake_initiator_iqn',
}

ISCSI_TARGET_DETAILS_LIST = [
    {'address': '5.6.7.8', 'port': '3260'},
    {'address': '1.2.3.4', 'port': '3260'},
    {'address': '99.98.97.96', 'port': '3260'},
]

IPV4_ADDRESS = '192.168.14.2'
NFS_SHARE_IPV4 = IPV4_ADDRESS + ':' + EXPORT_PATH

RESERVED_PERCENTAGE = 7
MAX_OVER_SUBSCRIPTION_RATIO = 19.0
TOTAL_BYTES = 4797892092432
AVAILABLE_BYTES = 13479932478
CAPACITY_VALUES = (TOTAL_BYTES, AVAILABLE_BYTES)
CAPACITIES = {'size-total': TOTAL_BYTES, 'size-available': AVAILABLE_BYTES}

IGROUP1 = {'initiator-group-os-type': 'linux',
           'initiator-group-type': 'fcp',
           'initiator-group-name': IGROUP1_NAME}

QOS_SPECS = {}
EXTRA_SPECS = {}
MAX_THROUGHPUT = '21734278B/s'
QOS_POLICY_GROUP_NAME = 'fake_qos_policy_group_name'

QOS_POLICY_GROUP_INFO_LEGACY = {
    'legacy': 'legacy-' + QOS_POLICY_GROUP_NAME,
    'spec': None,
}

QOS_POLICY_GROUP_SPEC = {
    'max_throughput': MAX_THROUGHPUT,
    'policy_name': QOS_POLICY_GROUP_NAME,
}

QOS_POLICY_GROUP_INFO = {'legacy': None, 'spec': QOS_POLICY_GROUP_SPEC}

CLONE_SOURCE_NAME = 'fake_clone_source_name'
CLONE_SOURCE_ID = 'fake_clone_source_id'
CLONE_SOURCE_SIZE = 1024

CLONE_SOURCE = {
    'size': CLONE_SOURCE_SIZE,
    'name': CLONE_SOURCE_NAME,
    'id': CLONE_SOURCE_ID,
}

CLONE_DESTINATION_NAME = 'fake_clone_destination_name'
CLONE_DESTINATION_SIZE = 1041
CLONE_DESTINATION_ID = 'fake_clone_destination_id'

CLONE_DESTINATION = {
    'size': CLONE_DESTINATION_SIZE,
    'name': CLONE_DESTINATION_NAME,
    'id': CLONE_DESTINATION_ID,
}

VOLUME_NAME = 'volume-fake_volume_id'
MOUNT_PATH = '168.10.16.11:/' + VOLUME_ID
SNAPSHOT_NAME = 'fake_snapshot_name'
SNAPSHOT_LUN_HANDLE = 'fake_snapshot_lun_handle'
SNAPSHOT_MOUNT = '/fake/mount/path'

SNAPSHOT = {
    'name': SNAPSHOT_NAME,
    'volume_size': SIZE,
    'volume_id': VOLUME_ID,
    'volume_name': VOLUME_NAME,
    'busy': False,
}

VOLUME_REF = {'name': 'fake_vref_name', 'size': 42}

FAKE_CMODE_VOLUMES = ['open123', 'mixed', 'open321']
FAKE_CMODE_POOL_MAP = {
    'open123': {
        'pool_name': 'open123',
    },
    'mixed': {
        'pool_name': 'mixed',
    },
    'open321': {
        'pool_name': 'open321',
    },
}

FILE_LIST = ['file1', 'file2', 'file3']

FAKE_LUN = netapp_api.NaElement.create_node_with_children(
    'lun-info',
    **{'alignment': 'indeterminate',
       'block-size': '512',
       'comment': '',
       'creation-timestamp': '1354536362',
       'is-space-alloc-enabled': 'false',
       'is-space-reservation-enabled': 'true',
       'mapped': 'false',
       'multiprotocol-type': 'linux',
       'online': 'true',
       'path': '/vol/fakeLUN/fakeLUN',
       'prefix-size': '0',
       'qtree': '',
       'read-only': 'false',
       'serial-number': '2FfGI$APyN68',
       'share-state': 'none',
       'size': '20971520',
       'size-used': '0',
       'staging': 'false',
       'suffix-size': '0',
       'uuid': 'cec1f3d7-3d41-11e2-9cf4-123478563412',
       'volume': 'fakeLUN',
       'vserver': 'fake_vserver'})

CG_VOLUME_NAME = 'fake_cg_volume'
CG_GROUP_NAME = 'fake_consistency_group'
CG_POOL_NAME = 'cdot'
SOURCE_CG_VOLUME_NAME = 'fake_source_cg_volume'
CG_VOLUME_ID = 'fake_cg_volume_id'
CG_VOLUME_SIZE = 100
SOURCE_CG_VOLUME_ID = 'fake_source_cg_volume_id'
CONSISTENCY_GROUP_NAME = 'fake_cg'
SOURCE_CONSISTENCY_GROUP_ID = 'fake_source_cg_id'
CONSISTENCY_GROUP_ID = 'fake_cg_id'
CG_SNAPSHOT_ID = 'fake_cg_snapshot_id'
CG_SNAPSHOT_NAME = 'snapshot-' + CG_SNAPSHOT_ID
CG_VOLUME_SNAPSHOT_ID = 'fake_cg_volume_snapshot_id'

CG_LUN_METADATA = {
    'OsType': None,
    'Path': '/vol/aggr1/fake_cg_volume',
    'SpaceReserved': 'true',
    'Qtree': None,
    'Volume': POOL_NAME,
}

SOURCE_CG_VOLUME = {
    'name': SOURCE_CG_VOLUME_NAME,
    'size': CG_VOLUME_SIZE,
    'id': SOURCE_CG_VOLUME_ID,
    'host': 'hostname@backend#cdot',
    'consistencygroup_id': None,
    'status': 'fake_status',
}

CG_VOLUME = {
    'name': CG_VOLUME_NAME,
    'size': 100,
    'id': CG_VOLUME_ID,
    'host': 'hostname@backend#' + CG_POOL_NAME,
    'consistencygroup_id': CONSISTENCY_GROUP_ID,
    'status': 'fake_status',
}

SOURCE_CONSISTENCY_GROUP = {
    'id': SOURCE_CONSISTENCY_GROUP_ID,
    'status': 'fake_status',
}

CONSISTENCY_GROUP = {
    'id': CONSISTENCY_GROUP_ID,
    'status': 'fake_status',
    'name': CG_GROUP_NAME,
}

CG_CONTEXT = {}

CG_SNAPSHOT = {
    'id': CG_SNAPSHOT_ID,
    'name': CG_SNAPSHOT_NAME,
    'volume_size': CG_VOLUME_SIZE,
    'consistencygroup_id': CONSISTENCY_GROUP_ID,
    'status': 'fake_status',
    'volume_id': 'fake_source_volume_id',
}

CG_VOLUME_SNAPSHOT = {
    'name': CG_SNAPSHOT_NAME,
    'volume_size': CG_VOLUME_SIZE,
    'cgsnapshot_id': CG_SNAPSHOT_ID,
    'id': CG_VOLUME_SNAPSHOT_ID,
    'status': 'fake_status',
    'volume_id': CG_VOLUME_ID,
}

SYSTEM_GET_VERSION_RESPONSE = etree.XML("""
  <results status="passed">
    <build-timestamp>1395426307</build-timestamp>
    <is-clustered>true</is-clustered>
    <version>NetApp Release 9.6P2: Fri Jul 19 06:06:59 UTC 2019</version>
    <version-tuple>
      <system-version-tuple>
        <generation>9</generation>
        <major>6</major>
        <minor>0</minor>
      </system-version-tuple>
    </version-tuple>
  </results>
""")


VG_VOLUME_NAME = 'fake_vg_volume'
VG_GROUP_NAME = 'fake_volume_group'
VG_POOL_NAME = 'cdot'
SOURCE_VG_VOLUME_NAME = 'fake_source_vg_volume'
VG_VOLUME_ID = 'fake_vg_volume_id'
VG_VOLUME_SIZE = 100
SOURCE_VG_VOLUME_ID = 'fake_source_vg_volume_id'
VOLUME_GROUP_NAME = 'fake_vg'
SOURCE_VOLUME_GROUP_ID = 'fake_source_vg_id'
VOLUME_GROUP_ID = 'fake_vg_id'
VG_SNAPSHOT_ID = 'fake_vg_snapshot_id'
VG_SNAPSHOT_NAME = 'snapshot-' + VG_SNAPSHOT_ID
VG_VOLUME_SNAPSHOT_ID = 'fake_vg_volume_snapshot_id'
MIN_SIZE_FOR_A_LUN = '4194304'
MAX_SIZE_FOR_A_LUN = '17555678822400'

VG_LUN_METADATA = {
    'OsType': None,
    'Path': '/vol/aggr1/fake_vg_volume',
    'SpaceReserved': 'true',
    'Qtree': None,
    'Volume': POOL_NAME,
}

SOURCE_VG_VOLUME = {
    'name': SOURCE_VG_VOLUME_NAME,
    'size': VG_VOLUME_SIZE,
    'id': SOURCE_VG_VOLUME_ID,
    'host': 'hostname@backend#cdot',
    'volumegroup_id': None,
    'status': 'fake_status',
    'provider_location': PROVIDER_LOCATION,
}

VG_VOLUME = {
    'name': VG_VOLUME_NAME,
    'size': 100,
    'id': VG_VOLUME_ID,
    'host': 'hostname@backend#' + VG_POOL_NAME,
    'volumegroup_id': VOLUME_GROUP_ID,
    'status': 'fake_status',
    'provider_location': PROVIDER_LOCATION,
}

SOURCE_VOLUME_GROUP = {
    'id': SOURCE_VOLUME_GROUP_ID,
    'status': 'fake_status',
}

VOLUME_GROUP = {
    'id': VOLUME_GROUP_ID,
    'status': 'fake_status',
    'name': VG_GROUP_NAME,
}

VG_CONTEXT = {}

VG_SNAPSHOT = {
    'id': VG_SNAPSHOT_ID,
    'name': VG_SNAPSHOT_NAME,
    'volume_size': VG_VOLUME_SIZE,
    'volumegroup_id': VOLUME_GROUP_ID,
    'status': 'fake_status',
    'volume_id': 'fake_source_volume_id',
    'volume': VG_VOLUME,
}

VG_VOLUME_SNAPSHOT = {
    'name': VG_SNAPSHOT_NAME,
    'volume_size': VG_VOLUME_SIZE,
    'vgsnapshot_id': VG_SNAPSHOT_ID,
    'id': VG_VOLUME_SNAPSHOT_ID,
    'status': 'fake_status',
    'volume_id': VG_VOLUME_ID,
}


class test_volume(object):

    def __getitem__(self, key):
        return getattr(self, key)


test_volume = test_volume()
test_volume.id = {'vserver': 'openstack', 'name': 'vola'}
test_volume.aggr = {
    'disk_type': 'SSD',
    'ha_policy': 'cfo',
    'junction': '/vola',
    'name': 'aggr1',
    'raid_type': 'raiddp',
}
test_volume.export = {'path': NFS_SHARE}
test_volume.sis = {'dedup': False, 'compression': False}
test_volume.state = {
    'status': 'online',
    'vserver_root': False,
    'junction_active': True,
}
test_volume.qos = {'qos_policy_group': None}
test_volume.host = 'fakehost@backbackend#fakepool'
test_volume.name = 'fakename'


class test_snapshot(object):
    pass

    def __getitem__(self, key):
        return getattr(self, key)


test_snapshot = test_snapshot()
test_snapshot.id = 'fake_snap_id'
test_snapshot.name = 'snapshot-%s' % test_snapshot.id
test_snapshot.volume_id = 'fake_volume_id'
test_snapshot.provider_location = PROVIDER_LOCATION


class test_iscsi_attachment(object):
    def __getattr__(self, key):
        return getattr(self, key)


test_iscsi_attachment = test_iscsi_attachment()
test_iscsi_attachment.connector = ISCSI_CONNECTOR


def get_fake_net_interface_get_iter_response():
    return etree.XML("""<results status="passed">
        <num-records>1</num-records>
        <attributes-list>
            <net-interface-info></net-interface-info>
            <address>FAKE_IP</address>
        </attributes-list>
    </results>""")


def get_fake_ifs():
    list_of_ifs = [
        etree.XML("""<net-interface-info>
        <address>FAKE_IP</address></net-interface-info>"""),
        etree.XML("""<net-interface-info>
        <address>FAKE_IP2</address></net-interface-info>"""),
        etree.XML("""<net-interface-info>
        <address>FAKE_IP3</address></net-interface-info>"""),
    ]
    return [netapp_api.NaElement(el) for el in list_of_ifs]



# Copyright (c) 2016 - 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import copy
import functools
import os
import random

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils

from cinder import exception
from cinder.i18n import _
from cinder.objects import fields
from cinder import utils as cinder_utils
from cinder.volume.drivers.dell_emc.unity import client
from cinder.volume.drivers.dell_emc.unity import utils
from cinder.volume import volume_types
from cinder.volume import volume_utils

storops = importutils.try_import('storops')
if storops:
    from storops import exception as storops_ex
    from storops.unity import enums
else:
    # Set storops_ex to be None for unit test
    storops_ex = None
    enums = None

LOG = logging.getLogger(__name__)

PROTOCOL_FC = 'FC'
PROTOCOL_ISCSI = 'iSCSI'


class VolumeParams(object):
    def __init__(self, adapter, volume, group_specs=None):
        self._adapter = adapter
        self._volume = volume

        self._volume_id = volume.id
        self._name = volume.name
        self._size = volume.size
        self._description = (volume.display_description
                             if volume.display_description
                             else volume.display_name)
        self._pool = None
        self._io_limit_policy = None
        self._is_thick = None
        self._is_compressed = None
        self._is_in_cg = None
        self._is_replication_enabled = None
        self._tiering_policy = None
        self.group_specs = group_specs if group_specs else {}

    @property
    def volume_id(self):
        return self._volume_id

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        self._size = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def pool(self):
        if self._pool is None:
            self._pool = self._adapter._get_target_pool(self._volume)
        return self._pool

    @pool.setter
    def pool(self, value):
        self._pool = value

    @property
    def io_limit_policy(self):
        if self._io_limit_policy is None:
            qos_specs = utils.get_backend_qos_specs(self._volume)
            self._io_limit_policy = self._adapter.client.get_io_limit_policy(
                qos_specs)
        return self._io_limit_policy

    @io_limit_policy.setter
    def io_limit_policy(self, value):
        self._io_limit_policy = value

    @property
    def is_thick(self):
        if self._is_thick is None:
            provision = utils.get_extra_spec(self._volume,
                                             utils.PROVISIONING_TYPE)
            support = utils.get_extra_spec(self._volume,
                                           'thick_provisioning_support')
            self._is_thick = (provision == 'thick' and support == '<is> True')
        return self._is_thick

    @property
    def is_compressed(self):
        if self._is_compressed is None:
            provision = utils.get_extra_spec(self._volume,
                                             utils.PROVISIONING_TYPE)
            compression = utils.get_extra_spec(self._volume,
                                               'compression_support')
            if (provision == utils.PROVISIONING_COMPRESSED and
                    compression == '<is> True'):
                self._is_compressed = True
        return self._is_compressed

    @is_compressed.setter
    def is_compressed(self, value):
        self._is_compressed = value

    @property
    def is_in_cg(self):
        if self._is_in_cg is None:
            self._is_in_cg = (self._volume.group and
                              volume_utils.is_group_a_cg_snapshot_type(
                                  self._volume.group))
        return self._is_in_cg

    @property
    def tiering_policy(self):
        tiering_policy_map = {'StartHighThenAuto':
                              enums.TieringPolicyEnum.AUTOTIER_HIGH,
                              'Auto':
                              enums.TieringPolicyEnum.AUTOTIER,
                              'HighestAvailable':
                              enums.TieringPolicyEnum.HIGHEST,
                              'LowestAvailable':
                              enums.TieringPolicyEnum.LOWEST}
        if not self._tiering_policy:
            tiering_value = utils.get_extra_spec(self._volume,
                                                 'storagetype:tiering')
            support = utils.get_extra_spec(self._volume,
                                           'fast_support') == '<is> True'

            if tiering_value and support:
                self._tiering_policy = tiering_policy_map.get(tiering_value)
            # if no value, unity sets StartHighThenAuto as default
        return self._tiering_policy

    @property
    def cg_id(self):
        if self.is_in_cg:
            return self._volume.group_id
        return None

    @property
    def is_replication_enabled(self):
        if self._is_replication_enabled is None:
            value = utils.get_extra_spec(self._volume, 'replication_enabled')
            self._is_replication_enabled = value == '<is> True'
        return self._is_replication_enabled

    def __eq__(self, other):
        return (self.volume_id == other.volume_id and
                self.name == other.name and
                self.size == other.size and
                self.io_limit_policy == other.io_limit_policy and
                self.is_thick == other.is_thick and
                self.is_compressed == other.is_compressed and
                self.is_in_cg == other.is_in_cg and
                self.cg_id == other.cg_id and
                self.tiering_policy == other.tiering_policy and
                self.is_replication_enabled == other.is_replication_enabled)


class CommonAdapter(object):
    protocol = 'unknown'
    driver_name = 'UnityAbstractDriver'
    driver_volume_type = 'unknown'

    def __init__(self, version=None):
        self.is_setup = False
        self.version = version
        self.driver = None
        self.config = None
        self.configured_pool_names = None
        self.reserved_percentage = None
        self.max_over_subscription_ratio = None
        self.volume_backend_name = None
        self.ip = None
        self.username = None
        self.password = None
        self.array_cert_verify = None
        self.array_ca_cert_path = None

        self._serial_number = None
        self.storage_pools_map = None
        self._client = None
        self.allowed_ports = None
        self.remove_empty_host = False
        self.to_lock_host = False
        self.replication_manager = None

    def do_setup(self, driver, conf):
        """Sets up the attributes of adapter.

        :param driver: the unity driver.
        :param conf: the driver configurations.
        """
        self.driver = driver
        self.config = self.normalize_config(conf)
        self.replication_manager = driver.replication_manager
        self.configured_pool_names = self.config.unity_storage_pool_names
        self.reserved_percentage = self.config.reserved_percentage
        self.max_over_subscription_ratio = (
            self.config.max_over_subscription_ratio)
        self.volume_backend_name = (self.config.safe_get('volume_backend_name')
                                    or self.driver_name)
        self.ip = self.config.san_ip
        self.username = self.config.san_login
        self.password = self.config.san_password
        # Allow for customized CA
        self.array_cert_verify = self.config.driver_ssl_cert_verify
        self.array_ca_cert_path = self.config.driver_ssl_cert_path

        sys_version = self.client.system.system_version
        if utils.is_before_4_1(sys_version):
            raise exception.VolumeBackendAPIException(
                data=_('Unity driver does not support array OE version: %s. '
                       'Upgrade to 4.1 or later.') % sys_version)

        self.storage_pools_map = self.get_managed_pools()

        self.allowed_ports = self.validate_ports(self.config.unity_io_ports)

        self.remove_empty_host = self.config.remove_empty_host
        self.to_lock_host = self.remove_empty_host

        group_name = (self.config.config_group if self.config.config_group
                      else 'DEFAULT')
        folder_name = '%(group)s.%(sys_name)s' % {
            'group': group_name, 'sys_name': self.client.system.info.name}
        persist_path = os.path.join(cfg.CONF.state_path, 'unity', folder_name)
        storops.TCHelper.set_up(persist_path)

        self.is_setup = True

    def normalize_config(self, config):
        config.unity_storage_pool_names = utils.remove_empty(
            '%s.unity_storage_pool_names' % config.config_group,
            config.unity_storage_pool_names)

        config.unity_io_ports = utils.remove_empty(
            '%s.unity_io_ports' % config.config_group,
            config.unity_io_ports)

        return config

    def get_all_ports(self):
        raise NotImplementedError()

    def validate_ports(self, ports_whitelist):
        all_ports = self.get_all_ports()
        # After normalize_config, `ports_whitelist` could be only None or valid
        # list in which the items are stripped.
        if ports_whitelist is None:
            return all_ports.id

        # For iSCSI port, the format is 'spa_eth0', and 'spa_iom_0_fc0' for FC.
        # Unix style glob like 'spa_*' is supported.
        whitelist = set(ports_whitelist)

        matched, _ignored, unmatched_whitelist = utils.match_any(all_ports.id,
                                                                 whitelist)
        if not matched:
            LOG.error('No matched ports filtered by all patterns: %s',
                      whitelist)
            raise exception.InvalidConfigurationValue(
                option='%s.unity_io_ports' % self.config.config_group,
                value=self.config.unity_io_ports)

        if unmatched_whitelist:
            LOG.error('No matched ports filtered by below patterns: %s',
                      unmatched_whitelist)
            raise exception.InvalidConfigurationValue(
                option='%s.unity_io_ports' % self.config.config_group,
                value=self.config.unity_io_ports)

        LOG.info('These ports %(matched)s will be used based on '
                 'the option unity_io_ports: %(config)s',
                 {'matched': matched,
                  'config': self.config.unity_io_ports})
        return matched

    @property
    def verify_cert(self):
        verify_cert = self.array_cert_verify
        if verify_cert and self.array_ca_cert_path is not None:
            verify_cert = self.array_ca_cert_path
        return verify_cert

    @property
    def client(self):
        if self._client is None:
            self._client = client.UnityClient(
                self.ip,
                self.username,
                self.password,
                verify_cert=self.verify_cert)
        return self._client

    @property
    def serial_number(self):
        if self._serial_number is None:
            self._serial_number = self.client.get_serial()
        return self._serial_number

    def get_managed_pools(self):
        names = self.configured_pool_names
        array_pools = self.client.get_pools()
        valid_names = utils.validate_pool_names(names, array_pools.name)
        return {p.name: p for p in array_pools if p.name in valid_names}

    def makeup_model(self, lun_id, is_snap_lun=False):
        lun_type = 'snap_lun' if is_snap_lun else 'lun'
        location = self._build_provider_location(lun_id=lun_id,
                                                 lun_type=lun_type)
        return {
            'provider_location': location,
            'provider_id': lun_id
        }

    def setup_replications(self, lun, model_update):
        if not self.replication_manager.is_replication_configured:
            LOG.debug('Replication device not configured, '
                      'skip setting up replication for lun %s',
                      lun.name)
            return model_update

        rep_data = {}
        rep_devices = self.replication_manager.replication_devices
        for backend_id, dst in rep_devices.items():
            remote_serial_number = dst.adapter.serial_number
            LOG.debug('Setting up replication to remote system %s',
                      remote_serial_number)
            remote_system = self.client.get_remote_system(remote_serial_number)
            if remote_system is None:
                raise exception.VolumeBackendAPIException(
                    data=_('Setup replication to remote system %s failed.'
                           'Cannot find it.') % remote_serial_number)
            rep_session = self.client.create_replication(
                lun, dst.max_time_out_of_sync,
                dst.destination_pool.get_id(), remote_system)
            rep_data[backend_id] = rep_session.name
        return utils.enable_replication_status(model_update, rep_data)

    def create_volume(self, volume):
        """Creates a volume.

        :param volume: volume information
        """
        params = VolumeParams(self, volume)
        log_params = {
            'name': params.name,
            'size': params.size,
            'description': params.description,
            'pool': params.pool,
            'io_limit_policy': params.io_limit_policy,
            'is_thick': params.is_thick,
            'is_compressed': params.is_compressed,
            'cg_id': params.cg_id,
            'is_replication_enabled': params.is_replication_enabled,
            'tiering_policy': params.tiering_policy
        }

        LOG.info('Create Volume: %(name)s, size: %(size)s, description: '
                 '%(description)s, pool: %(pool)s, io limit policy: '
                 '%(io_limit_policy)s, thick: %(is_thick)s, '
                 'compressed: %(is_compressed)s, cg_group: %(cg_id)s, '
                 'replication_enabled: %(is_replication_enabled)s.',
                 log_params)

        lun = self.client.create_lun(
            name=params.name,
            size=params.size,
            pool=params.pool,
            description=params.description,
            io_limit_policy=params.io_limit_policy,
            is_thin=False if params.is_thick else None,
            is_compressed=params.is_compressed,
            tiering_policy=params.tiering_policy)
        if params.cg_id:
            if self.client.is_cg_replicated(params.cg_id):
                msg = (_('Consistency group %(cg_id)s is in '
                         'replication status, cannot add lun to it.')
                       % {'cg_id': params.cg_id})
                raise exception.InvalidGroupStatus(reason=msg)
            LOG.info('Adding lun %(lun)s to cg %(cg)s.',
                     {'lun': lun.get_id(), 'cg': params.cg_id})
            self.client.update_cg(params.cg_id, [lun.get_id()], ())

        model_update = self.makeup_model(lun.get_id())

        if params.is_replication_enabled:
            if not params.cg_id:
                model_update = self.setup_replications(
                    lun, model_update)
            else:
                # Volume replication_status need be disabled
                # And be controlled by group replication
                model_update['replication_status'] = (
                    fields.ReplicationStatus.DISABLED)
        return model_update

    def delete_volume(self, volume):
        lun_id = self.get_lun_id(volume)
        if lun_id is None:
            LOG.info('Backend LUN not found, skipping the deletion. '
                     'Volume: %(volume_name)s.',
                     {'volume_name': volume.name})
        else:
            self.client.delete_lun(lun_id)

    def retype(self, ctxt, volume, new_type, diff, host):
        """Changes volume from one type to another."""
        old_qos_specs = {utils.QOS_SPECS: None}
        old_provision = None
        new_specs = volume_types.get_volume_type_extra_specs(
            new_type.get(utils.QOS_ID))
        new_qos_specs = volume_types.get_volume_type_qos_specs(
            new_type.get(utils.QOS_ID))
        lun = self.client.get_lun(name=volume.name)
        volume_type_id = volume.volume_type_id
        if volume_type_id:
            old_provision = utils.get_extra_spec(volume,
                                                 utils.PROVISIONING_TYPE)
            old_qos_specs = volume_types.get_volume_type_qos_specs(
                volume_type_id)

        need_migration = utils.retype_need_migration(
            volume, old_provision,
            new_specs.get(utils.PROVISIONING_TYPE), host)
        need_change_compress = utils.retype_need_change_compression(
            old_provision, new_specs.get(utils.PROVISIONING_TYPE))
        need_change_qos = utils.retype_need_change_qos(
            old_qos_specs, new_qos_specs)

        if need_migration or need_change_compress[0] or need_change_qos:
            if self.client.lun_has_snapshot(lun):
                LOG.warning('Driver is not able to do retype because '
                            'the volume %s has snapshot(s).',
                            volume.id)
                return False

        new_qos_dict = new_qos_specs.get(utils.QOS_SPECS)
        if need_change_qos:
            new_io_policy = (self.client.get_io_limit_policy(new_qos_dict)
                             if need_change_qos else None)
            # Modify lun to change qos settings
            if new_io_policy:
                lun.modify(io_limit_policy=new_io_policy)
            else:
                # remove current qos settings
                old_qos_dict = old_qos_specs.get(utils.QOS_SPECS)
                old_io_policy = self.client.get_io_limit_policy(old_qos_dict)
                old_io_policy.remove_from_storage(lun)

        if need_migration:
            LOG.debug('Driver needs to use storage-assisted migration '
                      'to retype the volume.')
            return self.migrate_volume(volume, host, new_specs)

        if need_change_compress[0]:
            # Modify lun to change compression
            lun.modify(is_compression=need_change_compress[1])
        return True

    def _create_host_and_attach(self, host_name, lun_or_snap):
        @utils.lock_if(self.to_lock_host, '{lock_name}')
        def _lock_helper(lock_name):
            if not self.to_lock_host:
                host = self.client.create_host(host_name)
            else:
                # Use the lock in the decorator
                host = self.client.create_host_wo_lock(host_name)
            hlu = self.client.attach(host, lun_or_snap)
            return host, hlu

        return _lock_helper('{unity}-{host}'.format(unity=self.client.host,
                                                    host=host_name))

    def _initialize_connection(self, lun_or_snap, connector, vol_id):
        host, hlu = self._create_host_and_attach(connector['host'],
                                                 lun_or_snap)
        self.client.update_host_initiators(
            host, self.get_connector_uids(connector))
        data = self.get_connection_info(hlu, host, connector)
        data['target_discovered'] = True
        if vol_id is not None:
            data['volume_id'] = vol_id
        conn_info = {
            'driver_volume_type': self.driver_volume_type,
            'data': data,
        }
        return conn_info

    @cinder_utils.trace
    def initialize_connection(self, volume, connector):
        lun = self.client.get_lun(lun_id=self.get_lun_id(volume))
        return self._initialize_connection(lun, connector, volume.id)

    @staticmethod
    def filter_targets_by_host(host):
        # No target info for iSCSI driver
        return []

    def _detach_and_delete_host(self, host_name, lun_or_snap,
                                is_multiattach_to_host=False):
        @utils.lock_if(self.to_lock_host, '{lock_name}')
        def _lock_helper(lock_name):
            # Only get the host from cache here
            host = self.client.create_host_wo_lock(host_name)
            if not is_multiattach_to_host:
                self.client.detach(host, lun_or_snap)
            host.update()  # need update to get the latest `host_luns`
            targets = self.filter_targets_by_host(host)
            if self.remove_empty_host and not host.host_luns:
                self.client.delete_host_wo_lock(host)
            return targets

        return _lock_helper('{unity}-{host}'.format(unity=self.client.host,
                                                    host=host_name))

    @staticmethod
    def get_terminate_connection_info(connector, targets):
        # No return data from terminate_connection for iSCSI driver
        return {}

    def _terminate_connection(self, lun_or_snap, connector,
                              is_multiattach_to_host=False):
        is_force_detach = connector is None
        data = {}
        if is_force_detach:
            self.client.detach_all(lun_or_snap)
        else:
            targets = self._detach_and_delete_host(
                connector['host'], lun_or_snap,
                is_multiattach_to_host=is_multiattach_to_host)
            data = self.get_terminate_connection_info(connector, targets)
        return {
            'driver_volume_type': self.driver_volume_type,
            'data': data,
        }

    @cinder_utils.trace
    def terminate_connection(self, volume, connector):
        lun = self.client.get_lun(lun_id=self.get_lun_id(volume))
        # None `connector` indicates force detach, then detach all even the
        # volume is multi-attached.
        multiattach_flag = (connector is not None and
                            utils.is_multiattach_to_host(
                                volume.volume_attachment,
                                connector['host']))
        return self._terminate_connection(
            lun, connector, is_multiattach_to_host=multiattach_flag)

    def get_connector_uids(self, connector):
        return None

    def get_connection_info(self, hlu, host, connector):
        return {}

    def extend_volume(self, volume, new_size):
        lun_id = self.get_lun_id(volume)
        if lun_id is None:
            msg = (_('Backend LUN not found for Volume: %(volume_name)s.') %
                   {'volume_name': volume.name})
            raise exception.VolumeBackendAPIException(data=msg)
        else:
            self.client.extend_lun(lun_id, new_size)

    def _get_target_pool(self, volume):
        return self.storage_pools_map[utils.get_pool_name(volume)]

    def _build_provider_location(self, lun_id=None, lun_type=None):
        return utils.build_provider_location(
            system=self.serial_number,
            lun_type=lun_type,
            lun_id=lun_id,
            version=self.version)

    @utils.append_capabilities
    def update_volume_stats(self):
        return {
            'volume_backend_name': self.volume_backend_name,
            'storage_protocol': self.protocol,
            'pools': self.get_pools_stats(),
            'replication_enabled':
                self.replication_manager.is_replication_configured,
            'replication_targets':
                list(self.replication_manager.replication_devices),
        }

    def get_pools_stats(self):
        self.storage_pools_map = self.get_managed_pools()
        return [self._get_pool_stats(pool) for pool in self.pools]

    @property
    def pools(self):
        return self.storage_pools_map.values()

    @utils.append_capabilities
    def _get_pool_stats(self, pool):
        return {
            'pool_name': pool.name,
            'total_capacity_gb': utils.byte_to_gib(pool.size_total),
            'provisioned_capacity_gb': utils.byte_to_gib(
                pool.size_subscribed),
            'free_capacity_gb': utils.byte_to_gib(pool.size_free),
            'reserved_percentage': self.reserved_percentage,
            'location_info': ('%(pool_name)s|%(array_serial)s' %
                              {'pool_name': pool.name,
                               'array_serial': self.serial_number}),
            'compression_support': pool.is_all_flash,
            'max_over_subscription_ratio': (
                self.max_over_subscription_ratio),
            'multiattach': True,
            'replication_enabled':
                self.replication_manager.is_replication_configured,
            'replication_targets':
                list(self.replication_manager.replication_devices),
        }

    def get_lun_id(self, volume):
        """Retrieves id of the volume's backing LUN.

        :param volume: volume information
        """
        if volume.provider_location:
            return utils.extract_provider_location(volume.provider_location,
                                                   'id')
        else:
            # In some cases, cinder will not update volume info in DB with
            # provider_location returned by us. We need to retrieve the id
            # from array.
            lun = self.client.get_lun(name=volume.name)
            return lun.get_id() if lun is not None else None

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: snapshot information.
        """
        src_lun_id = self.get_lun_id(snapshot.volume)
        snap = self.client.create_snap(src_lun_id, snapshot.name)
        location = self._build_provider_location(lun_type='snapshot',
                                                 lun_id=snap.get_id())
        return {'provider_location': location,
                'provider_id': snap.get_id()}

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: the snapshot to delete.
        """
        snap = self.client.get_snap(name=snapshot.name)
        self.client.delete_snap(snap)

    def _get_referenced_lun(self, existing_ref):
        if 'source-id' in existing_ref:
            lun = self.client.get_lun(lun_id=existing_ref['source-id'])
        elif 'source-name' in existing_ref:
            lun = self.client.get_lun(name=existing_ref['source-name'])
        else:
            reason = _('Reference must contain source-id or source-name key.')
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=reason)
        if lun is None or not lun.existed:
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=_("LUN doesn't exist."))
        return lun

    def manage_existing(self, volume, existing_ref):
        """Manages an existing LUN in the array.

        The LUN should be in a manageable pool backend, otherwise return error.
        Rename the backend storage object so that it matches the
        `volume['name']` which is how drivers traditionally map between a
        cinder volume and the associated backend storage object.

        LUN ID or name are supported in `existing_ref`, like:

        .. code-block:: none

        existing_ref:{

            'source-id':<LUN id in Unity>

        }

        or

        .. code-block:: none

        existing_ref:{

            'source-name':<LUN name in Unity>

        }
        """
        lun = self._get_referenced_lun(existing_ref)
        lun.modify(name=volume.name)
        return {
            'provider_location':
                self._build_provider_location(lun_id=lun.get_id(),
                                              lun_type='lun'),
            'provider_id': lun.get_id()
        }

    def manage_existing_get_size(self, volume, existing_ref):
        """Returns size of volume to be managed by `manage_existing`.

        The driver does some check here:
        1. The LUN `existing_ref` should be managed by the `volume.host`.
        """
        lun = self._get_referenced_lun(existing_ref)
        target_pool_name = utils.get_pool_name(volume)
        lun_pool_name = lun.pool.name
        if target_pool_name and lun_pool_name != target_pool_name:
            reason = (_('The imported LUN is in pool %(pool_name)s '
                        'which is not managed by the host %(host)s.') %
                      {'pool_name': lun_pool_name,
                       'host': volume.host})
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=reason)

        return utils.byte_to_gib(lun.size_total)

    def _disconnect_device(self, conn):
        conn['connector'].disconnect_volume(conn['conn']['data'],
                                            conn['device'])

    def _connect_device(self, conn):
        return self.driver._connect_device(conn)

    @contextlib.contextmanager
    def _connect_resource(self, lun_or_snap, connector, res_id):
        """Connects to LUN or snapshot, and makes sure disconnect finally.

        :param lun_or_snap: the LUN or snapshot to connect/disconnect.
        :param connector: the host connector information.
        :param res_id: the ID of the LUN or snapshot.

        :return: the connection information, in a dict with format
         like (same as the one returned by `_connect_device`):
         {
            'conn': <info returned by `initialize_connection`>,
            'device': <value returned by `connect_volume`>,
            'connector': <host connector info>
         }
        """
        init_conn_func = functools.partial(self._initialize_connection,
                                           lun_or_snap, connector, res_id)
        term_conn_func = functools.partial(self._terminate_connection,
                                           lun_or_snap, connector)
        with utils.assure_cleanup(init_conn_func, term_conn_func,
                                  False) as conn_info:
            conn_device_func = functools.partial(self._connect_device,
                                                 conn_info)
            with utils.assure_cleanup(conn_device_func,
                                      self._disconnect_device,
                                      True) as attach_info:
                yield attach_info

    def _dd_copy(self, vol_params, src_snap, src_lun=None):
        """Creates a volume via copying a Unity snapshot.

        It attaches the `volume` and `snap`, then use `dd` to copy the
        data from the Unity snapshot to the `volume`.
        """
        dest_lun = self.client.create_lun(
            name=vol_params.name, size=vol_params.size, pool=vol_params.pool,
            description=vol_params.description,
            io_limit_policy=vol_params.io_limit_policy,
            is_thin=False if vol_params.is_thick else None,
            is_compressed=vol_params.is_compressed)
        src_id = src_snap.get_id()
        try:
            conn_props = cinder_utils.brick_get_connector_properties()

            with self._connect_resource(dest_lun, conn_props,
                                        vol_params.volume_id) as dest_info, \
                    self._connect_resource(src_snap, conn_props,
                                           src_id) as src_info:
                if src_lun is None:
                    # If size is not specified, need to get the size from LUN
                    # of snapshot.
                    size_in_m = utils.byte_to_mib(src_snap.size)
                else:
                    size_in_m = utils.byte_to_mib(src_lun.size_total)
                volume_utils.copy_volume(
                    src_info['device']['path'],
                    dest_info['device']['path'],
                    size_in_m,
                    self.driver.configuration.volume_dd_blocksize,
                    sparse=True)
        except Exception:
            with excutils.save_and_reraise_exception():
                utils.ignore_exception(self.client.delete_lun,
                                       dest_lun.get_id())
                LOG.error('Failed to create cloned volume: %(vol_id)s, '
                          'from source unity snapshot: %(snap_name)s.',
                          {'vol_id': vol_params.volume_id,
                           'snap_name': src_snap.name})

        return dest_lun

    def _thin_clone(self, vol_params, src_snap, src_lun=None):
        tc_src = src_snap if src_lun is None else src_lun
        try:
            LOG.debug('Try to thin clone from %s.', tc_src.name)
            lun = self.client.thin_clone(
                tc_src, vol_params.name,
                description=vol_params.description,
                io_limit_policy=vol_params.io_limit_policy,
                new_size_gb=vol_params.size)
        except storops_ex.UnityThinCloneLimitExceededError:
            LOG.info('Number of thin clones of base LUN exceeds system '
                     'limit, dd-copy a new one and thin clone from it.')
            # Copy via dd if thin clone meets the system limit
            hidden = copy.copy(vol_params)
            hidden.name = 'hidden-%s' % vol_params.name
            hidden.description = 'hidden-%s' % vol_params.description
            copied_lun = self._dd_copy(hidden, src_snap, src_lun=src_lun)
            LOG.debug('Notify storops the dd action of lun: %(src_name)s. And '
                      'the newly copied lun is: %(copied)s.',
                      {'src_name': tc_src.name, 'copied': copied_lun.name})
            storops.TCHelper.notify(tc_src,
                                    storops.ThinCloneActionEnum.DD_COPY,
                                    copied_lun)
            lun = self.client.thin_clone(
                copied_lun, vol_params.name,
                description=vol_params.description,
                io_limit_policy=vol_params.io_limit_policy,
                new_size_gb=vol_params.size)
        except storops_ex.SystemAPINotSupported:
            # Thin clone not support on array version before Merlin
            lun = self._dd_copy(vol_params, src_snap, src_lun=src_lun)
            LOG.debug(
                'Volume copied via dd because array OE is too old to support '
                'thin clone api. source snap: %(src_snap)s, lun: %(src_lun)s.',
                {'src_snap': src_snap.name,
                 'src_lun': 'Unknown' if src_lun is None else src_lun.name})
        except storops_ex.UnityThinCloneNotAllowedError:
            # Thin clone not allowed on some resources,
            # like thick luns and their snaps
            lun = self._dd_copy(vol_params, src_snap, src_lun=src_lun)
            LOG.debug(
                'Volume copied via dd because source snap/lun is not allowed '
                'to thin clone, i.e. it is thick. source snap: %(src_snap)s, '
                'lun: %(src_lun)s.',
                {'src_snap': src_snap.name,
                 'src_lun': 'Unknown' if src_lun is None else src_lun.name})
        return lun

    def create_volume_from_snapshot(self, volume, snapshot):
        snap = self.client.get_snap(snapshot.name)
        params = VolumeParams(self, volume)
        lun = self._thin_clone(params, snap)
        model_update = self.makeup_model(lun.get_id(), is_snap_lun=True)

        if params.is_replication_enabled:
            model_update = self.setup_replications(lun, model_update)
        return model_update

    def create_cloned_volume(self, volume, src_vref):
        """Creates cloned volume.

        1. Take an internal snapshot of source volume, and attach it.
        2. Thin clone from the snapshot to a new volume.
           Note: there are several cases the thin clone will downgrade to `dd`,
           2.1 Source volume is attached (in-use).
           2.2 Array OE version doesn't support thin clone.
           2.3 The current LUN family reaches the thin clone limits.
        3. Delete the internal snapshot created in step 1.
        """

        src_lun_id = self.get_lun_id(src_vref)
        if src_lun_id is None:
            raise exception.VolumeBackendAPIException(
                data=_(
                    "LUN ID of source volume: %s not found.") % src_vref.name)
        src_lun = self.client.get_lun(lun_id=src_lun_id)
        src_snap_name = 'snap_clone_%s' % volume.id

        create_snap_func = functools.partial(self.client.create_snap,
                                             src_lun_id, src_snap_name)
        vol_params = VolumeParams(self, volume)
        with utils.assure_cleanup(create_snap_func,
                                  self.client.delete_snap,
                                  True) as src_snap:
            LOG.debug('Internal snapshot for clone is created, '
                      'name: %(name)s, id: %(id)s.',
                      {'name': src_snap_name,
                       'id': src_snap.get_id()})
            if src_vref.volume_attachment:
                lun = self._dd_copy(vol_params, src_snap, src_lun=src_lun)
                LOG.debug('Volume copied using dd because source volume: '
                          '%(name)s is attached: %(attach)s.',
                          {'name': src_vref.name,
                           'attach': src_vref.volume_attachment})
                model_update = self.makeup_model(lun.get_id())
            else:
                lun = self._thin_clone(vol_params, src_snap, src_lun=src_lun)
                model_update = self.makeup_model(lun.get_id(),
                                                 is_snap_lun=True)

            if vol_params.is_replication_enabled:
                model_update = self.setup_replications(lun, model_update)
            return model_update

    def get_pool_name(self, volume):
        return self.client.get_pool_name(volume.name)

    def get_pool_id_by_name(self, name):
        return self.client.get_pool_id_by_name(name=name)

    @cinder_utils.trace
    def initialize_connection_snapshot(self, snapshot, connector):
        snap = self.client.get_snap(snapshot.name)
        return self._initialize_connection(snap, connector, snapshot.id)

    @cinder_utils.trace
    def terminate_connection_snapshot(self, snapshot, connector):
        snap = self.client.get_snap(snapshot.name)
        return self._terminate_connection(snap, connector)

    @cinder_utils.trace
    def restore_snapshot(self, volume, snapshot):
        return self.client.restore_snapshot(snapshot.name)

    def migrate_volume(self, volume, host, extra_specs=None):
        """Leverage the Unity move session functionality.

        This method is invoked at the source backend.

        :param extra_specs: Instance of ExtraSpecs. The new volume will be
            changed to align with the new extra specs.
        """
        log_params = {
            'name': volume.name,
            'src_host': volume.host,
            'dest_host': host['host'],
            'extra_specs': extra_specs,
        }
        LOG.info('Migrate Volume: %(name)s, host: %(src_host)s, destination: '
                 '%(dest_host)s, extra_specs: %(extra_specs)s', log_params)

        src_backend = utils.get_backend_name_from_volume(volume)
        dest_backend = utils.get_backend_name_from_host(host)

        if src_backend != dest_backend:
            LOG.debug('Cross-backends migration not supported by Unity '
                      'driver. Falling back to host-assisted migration.')
            return False, None

        lun_id = self.get_lun_id(volume)
        provision = None
        if extra_specs:
            provision = extra_specs.get(utils.PROVISIONING_TYPE)
        dest_pool_name = utils.get_pool_name_from_host(host)
        dest_pool_id = self.get_pool_id_by_name(dest_pool_name)
        if self.client.migrate_lun(lun_id, dest_pool_id, provision):
            LOG.debug('Volume migrated successfully.')
            model_update = {}
            return True, model_update

        LOG.debug('Volume migrated failed. Falling back to '
                  'host-assisted migration.')
        return False, None

    def create_group(self, group):
        """Creates a generic group.

        :param group: group information
        """
        cg_name = group.id
        description = group.description if group.description else group.name

        LOG.info('Create group: %(name)s, description: %(description)s',
                 {'name': cg_name, 'description': description})

        self.client.create_cg(cg_name, description=description)
        return {'status': fields.GroupStatus.AVAILABLE}

    def delete_group(self, group):
        """Deletes the generic group.

        :param group: the group to delete
        """

        # Deleting cg will also delete all the luns in it.
        group_id = group.id
        if self.client.is_cg_replicated(group_id):
            self.client.delete_cg_rep_session(group_id)
        self.client.delete_cg(group_id)
        return None, None

    def update_group(self, group, add_volumes, remove_volumes):
        add_lun_ids = (set(map(self.get_lun_id, add_volumes)) if add_volumes
                       else set())
        remove_lun_ids = (set(map(self.get_lun_id, remove_volumes))
                          if remove_volumes else set())
        self.client.update_cg(group.id, add_lun_ids, remove_lun_ids)
        return {'status': fields.GroupStatus.AVAILABLE}, None, None

    def copy_luns_in_group(self, group, volumes, src_cg_snap, src_volumes):
        # Use dd to copy data here. The reason why not using thinclone is:
        # 1. Cannot use cg thinclone due to the tight couple between source
        # group and cloned one.
        # 2. Cannot use lun thinclone due to clone lun in cg is not supported.

        lun_snaps = self.client.filter_snaps_in_cg_snap(src_cg_snap.id)

        # Make sure the `lun_snaps` is as order of `src_volumes`
        src_lun_ids = [self.get_lun_id(volume) for volume in src_volumes]
        lun_snaps.sort(key=lambda snap: src_lun_ids.index(snap.lun.id))

        dest_luns = [self._dd_copy(VolumeParams(self, dest_volume), lun_snap)
                     for dest_volume, lun_snap in zip(volumes, lun_snaps)]

        self.client.create_cg(group.id, lun_add=dest_luns)
        return ({'status': fields.GroupStatus.AVAILABLE},
                [{'id': dest_volume.id, 'status': fields.GroupStatus.AVAILABLE}
                 for dest_volume in volumes])

    def create_group_from_snap(self, group, volumes,
                               group_snapshot, snapshots):
        src_cg_snap = self.client.get_snap(group_snapshot.id)
        src_vols = ([snap.volume for snap in snapshots] if snapshots else [])
        return self.copy_luns_in_group(group, volumes, src_cg_snap, src_vols)

    def create_cloned_group(self, group, volumes, source_group, source_vols):
        src_group_snap_name = 'snap_clone_group_{}'.format(source_group.id)
        create_snap_func = functools.partial(self.client.create_cg_snap,
                                             source_group.id,
                                             src_group_snap_name)
        with utils.assure_cleanup(create_snap_func,
                                  self.client.delete_snap,
                                  True) as src_cg_snap:
            LOG.debug('Internal group snapshot for clone is created, '
                      'name: %(name)s, id: %(id)s.',
                      {'name': src_group_snap_name,
                       'id': src_cg_snap.get_id()})
            source_vols = source_vols if source_vols else []
            return self.copy_luns_in_group(group, volumes, src_cg_snap,
                                           source_vols)

    def create_group_snapshot(self, group_snapshot, snapshots):
        self.client.create_cg_snap(group_snapshot.group_id,
                                   snap_name=group_snapshot.id)

        model_update = {'status': fields.GroupStatus.AVAILABLE}
        snapshots_model_update = [{'id': snapshot.id,
                                   'status': fields.SnapshotStatus.AVAILABLE}
                                  for snapshot in snapshots]
        return model_update, snapshots_model_update

    def delete_group_snapshot(self, group_snapshot):
        cg_snap = self.client.get_snap(group_snapshot.id)
        self.client.delete_snap(cg_snap)
        return None, None

    def enable_replication(self, context, group, volumes):
        """Enable the group replication."""

        @cinder_utils.retry(exception.InvalidGroup, interval=20, retries=6)
        def _wait_until_cg_not_replicated(_client, _cg_id):
            cg = _client.get_cg(name=_cg_id)
            if cg.check_cg_is_replicated():
                msg = _('The remote cg (%s) is still in replication status, '
                        'maybe the source cg was just deleted, '
                        'retrying.') % group_id
                LOG.info(msg)
                raise exception.InvalidGroup(reason=msg)

            return cg

        group_update = {}
        group_id = group.id
        if not volumes:
            LOG.warning('There is no Volume in group: %s, cannot enable '
                        'group replication', group_id)
            return group_update, []
        # check whether the group was created as cg in unity
        group_is_cg = utils.group_is_cg(group)
        if not group_is_cg:
            msg = (_('Cannot enable replication on generic group '
                     '%(group_id)s, need to use CG type instead '
                     '(need to enable consistent_group_snapshot_enabled in '
                     'the group type).')
                   % {'group_id': group_id})
            raise exception.InvalidGroupType(reason=msg)

        cg = self.client.get_cg(name=group_id)
        try:
            if not cg.check_cg_is_replicated():
                rep_devices = self.replication_manager.replication_devices
                for backend_id, dst in rep_devices.items():
                    remote_serial_number = dst.adapter.serial_number
                    max_time = dst.max_time_out_of_sync
                    pool_id = dst.destination_pool.get_id()
                    _client = dst.adapter.client
                    remote_system = self.client.get_remote_system(
                        remote_serial_number)
                    # check if remote cg exists and delete it
                    # before enable replication
                    remote_cg = _wait_until_cg_not_replicated(_client,
                                                              group_id)
                    remote_cg.delete()
                    # create cg replication session
                    self.client.create_cg_replication(
                        group_id, pool_id, remote_system, max_time)
                    group_update.update({
                        'replication_status':
                            fields.ReplicationStatus.ENABLED})
            else:
                LOG.info('group: %s is already in replication, no need to '
                         'enable again.', group_id)
        except Exception as e:
            group_update.update({
                'replication_status': fields.ReplicationStatus.ERROR})
            LOG.error("Error enabling replication on group %(group)s. "
                      "Exception received: %(e)s.",
                      {'group': group.id, 'e': e})
        return group_update, None

    def disable_replication(self, context, group, volumes):
        """Disable the group replication."""
        group_update = {}
        group_id = group.id
        if not volumes:
            # Return if empty group
            LOG.warning('There is no Volume in group: %s, cannot disable '
                        'group replication', group_id)
            return group_update, []
        group_is_cg = utils.group_is_cg(group)
        if not group_is_cg:
            msg = (_('Cannot disable replication on generic group '
                     '%(group_id)s, need use CG type instead of '
                     'that (need enable '
                     'consistent_group_snapshot_enabled in '
                     'group type).')
                   % {'group_id': group_id})
            raise exception.InvalidGroupType(reason=msg)
        try:
            if self.client.is_cg_replicated(group_id):
                # delete rep session if exists
                self.client.delete_cg_rep_session(group_id)
            if not self.client.is_cg_replicated(group_id):
                LOG.info('Group is not in replication, '
                         'not need to disable replication again.')

            group_update.update({
                'replication_status': fields.ReplicationStatus.DISABLED})
        except Exception as e:
            group_update.update({
                'replication_status': fields.ReplicationStatus.ERROR})
            LOG.error("Error disabling replication on group %(group)s. "
                      "Exception received: %(e)s.",
                      {'group': group.id, 'e': e})
        return group_update, None

    def failover_replication(self, context, group, volumes,
                             secondary_id):
        """"Fail-over the consistent group."""
        group_update = {}
        volume_update_list = []
        if not volumes:
            # Return if empty group
            return group_update, volume_update_list

        group_is_cg = utils.group_is_cg(group)
        group_id = group.id
        if not group_is_cg:
            msg = (_('Cannot failover replication on generic group '
                     '%(group_id)s, need use CG type instead of '
                     'that (need enable '
                     'consistent_group_snapshot_enabled in '
                     'group type).')
                   % {'group_id': group_id})
            raise exception.InvalidGroupType(reason=msg)

        real_secondary_id = random.choice(
            list(self.replication_manager.replication_devices))

        group_update = {'replication_status': group.replication_status}
        if self.client.is_cg_replicated(group_id):
            try:
                if secondary_id != 'default':
                    try:
                        # Planed failover after sync date when the source unity
                        # is in health status
                        self.client.failover_cg_rep_session(group_id, True)
                    except Exception as ex:
                        LOG.warning('ERROR happened when failover from source '
                                    'unity, issue details: %s. Try failover '
                                    'from target unity', ex)
                        # Something wrong with the source unity, try failover
                        # from target unity without sync date
                        _adapter = self.replication_manager.replication_devices
                        [real_secondary_id].adapter
                        _client = _adapter.client
                        _client.failover_cg_rep_session(group_id, False)
                    rep_status = fields.ReplicationStatus.FAILED_OVER
                else:
                    # start failback when secondary_id is 'default'
                    _adapter = self.replication_manager.replication_devices[
                        real_secondary_id].adapter
                    _client = _adapter.client
                    _client.failback_cg_rep_session(group_id)
                    rep_status = fields.ReplicationStatus.ENABLED
            except Exception as ex:
                rep_status = fields.ReplicationStatus.ERROR
                LOG.error("Error failover replication on group %(group)s. "
                          "Exception received: %(e)s.",
                          {'group': group_id, 'e': ex})

            group_update['replication_status'] = rep_status
            for volume in volumes:
                volume_update = {
                    'id': volume.id,
                    'replication_status': rep_status}
                volume_update_list.append(volume_update)
        return group_update, volume_update_list

    def get_replication_error_status(self, context, groups):
        """The failover only happens manually, no need to update the status."""
        return [], []

    @cinder_utils.trace
    def failover(self, volumes, secondary_id=None, groups=None):
        # TODO(ryan) support group failover after group bp merges
        # https://review.opendev.org/#/c/574119/

        if secondary_id is None:
            LOG.debug('No secondary specified when failover. '
                      'Randomly choose a secondary')
            secondary_id = random.choice(
                list(self.replication_manager.replication_devices))
            LOG.debug('Chose %s as secondary', secondary_id)

        is_failback = secondary_id == 'default'

        def _failover_or_back(volume):
            LOG.debug('Failing over volume: %(vol)s to secondary id: '
                      '%(sec_id)s', vol=volume.name, sec_id=secondary_id)
            model_update = {
                'volume_id': volume.id,
                'updates': {}
            }

            if not volume.replication_driver_data:
                LOG.error('Empty replication_driver_data of volume: %s, '
                          'replication session name should be in it.',
                          volume.name)
                return utils.error_replication_status(model_update)
            rep_data = utils.load_replication_data(
                volume.replication_driver_data)

            if is_failback:
                # Failback executed on secondary backend which is currently
                # active.
                _adapter = self.replication_manager.default_device.adapter
                _client = self.replication_manager.active_adapter.client
                rep_name = rep_data[self.replication_manager.active_backend_id]
            else:
                # Failover executed on secondary backend because primary could
                # die.
                _adapter = self.replication_manager.replication_devices[
                    secondary_id].adapter
                _client = _adapter.client
                rep_name = rep_data[secondary_id]

            try:
                rep_session = _client.get_replication_session(name=rep_name)

                if is_failback:
                    _client.failback_replication(rep_session)
                    new_model = _adapter.makeup_model(
                        rep_session.src_resource_id)
                else:
                    _client.failover_replication(rep_session)
                    new_model = _adapter.makeup_model(
                        rep_session.dst_resource_id)

                model_update['updates'].update(new_model)
                self.replication_manager.failover_service(secondary_id)
                return model_update
            except client.ClientReplicationError as ex:
                LOG.error('Failover failed, volume: %(vol)s, secondary id: '
                          '%(sec_id)s, error: %(err)s',
                          vol=volume.name, sec_id=secondary_id, err=ex)
                return utils.error_replication_status(model_update)

        return (secondary_id,
                [_failover_or_back(volume) for volume in volumes],
                [])


class ISCSIAdapter(CommonAdapter):
    protocol = PROTOCOL_ISCSI
    driver_name = 'UnityISCSIDriver'
    driver_volume_type = 'iscsi'

    def get_all_ports(self):
        return self.client.get_ethernet_ports()

    def get_connector_uids(self, connector):
        return utils.extract_iscsi_uids(connector)

    def get_connection_info(self, hlu, host, connector):
        targets = self.client.get_iscsi_target_info(self.allowed_ports)
        if not targets:
            msg = _("There is no accessible iSCSI targets on the system.")
            raise exception.VolumeBackendAPIException(data=msg)
        one_target = random.choice(targets)
        portals = [a['portal'] for a in targets]
        iqns = [a['iqn'] for a in targets]
        data = {
            'target_luns': [hlu] * len(portals),
            'target_iqns': iqns,
            'target_portals': portals,
            'target_lun': hlu,
            'target_portal': one_target['portal'],
            'target_iqn': one_target['iqn'],
        }
        return data


class FCAdapter(CommonAdapter):
    protocol = PROTOCOL_FC
    driver_name = 'UnityFCDriver'
    driver_volume_type = 'fibre_channel'

    def __init__(self, version=None):
        super(FCAdapter, self).__init__(version=version)
        self.lookup_service = None

    def do_setup(self, driver, config):
        super(FCAdapter, self).do_setup(driver, config)
        self.lookup_service = utils.create_lookup_service()

    def get_all_ports(self):
        return self.client.get_fc_ports()

    def get_connector_uids(self, connector):
        return utils.extract_fc_uids(connector)

    @property
    def auto_zone_enabled(self):
        return self.lookup_service is not None

    def get_connection_info(self, hlu, host, connector):
        targets = self.client.get_fc_target_info(
            host, logged_in_only=(not self.auto_zone_enabled),
            allowed_ports=self.allowed_ports)

        if not targets:
            msg = _("There is no accessible fibre channel targets on the "
                    "system.")
            raise exception.VolumeBackendAPIException(data=msg)

        if self.auto_zone_enabled:
            data = self._get_fc_zone_info(connector['wwpns'], targets)
        else:
            data = {
                'target_wwn': targets,
            }
        data['target_lun'] = hlu
        return data

    def filter_targets_by_host(self, host):
        if self.auto_zone_enabled and not host.host_luns:
            return self.client.get_fc_target_info(
                host=host, logged_in_only=False,
                allowed_ports=self.allowed_ports)
        return []

    def get_terminate_connection_info(self, connector, targets):
        # For FC, terminate_connection needs to return data to zone manager
        # which would clean the zone based on the data.
        if targets:
            return self._get_fc_zone_info(connector['wwpns'], targets)
        return {}

    def _get_fc_zone_info(self, initiator_wwns, target_wwns):
        mapping = self.lookup_service.get_device_mapping_from_network(
            initiator_wwns, target_wwns)
        targets, itor_tgt_map = utils.convert_to_itor_tgt_map(mapping)
        return {
            'target_wwn': targets,
            'initiator_target_map': itor_tgt_map,
        }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import pprint
import re

from oslo_serialization import jsonutils
import six

from cinder.tests.functional import functional_helpers
from cinder.tests.unit import test

PROJECT_ID = "6f70656e737461636b20342065766572"

# for pretty printing errors
pp = pprint.PrettyPrinter(indent=4)


class NoMatch(test.TestingException):
    pass


def pretty_data(data):
    data = jsonutils.dumps(jsonutils.loads(data), sort_keys=True,
                           indent=4)
    return '\n'.join(line.rstrip() for line in data.split('\n')).strip()


def objectify(data):
    if not data:
        return {}
    # templates will contain values like %(foo)s
    # throughout them. If these are inside of double quoted
    # strings, life is good, and we can treat it just like valid
    # json to load it to python.
    #
    # However we've got some fields which are ints, like
    # aggregate_id. This means we've got a snippet in the sample
    # that looks like:
    #
    #     "id": %(aggregate_id)s,
    #
    # which is not valid json, and will explode. We do a quick and
    # dirty transform of this to:
    #
    #     "id": "%(int:aggregate_id)s",
    #
    # That makes it valid data to convert to json, but keeps
    # around the information that we need to drop those strings
    # later. The regex anchors from the ': ', as all of these will
    # be top rooted keys.
    data = re.sub(r'(\: )%\((.+)\)s([^"])', r'\1"%(int:\2)s"\3', data)
    return jsonutils.loads(data)


class ApiSampleTestBase(functional_helpers._FunctionalTestBase):
    all_extensions = True
    sample_dir = None
    _project_id = True
    _use_common_volume_api_samples = False

    def __init__(self, *args, **kwargs):
        super(ApiSampleTestBase, self).__init__(*args, **kwargs)
        self.subs = {}

    def setUp(self):
        super(ApiSampleTestBase, self).setUp()
        self.api_major_version = 'v3'

        # this is used to generate sample docs
        self.generate_samples = os.getenv('GENERATE_SAMPLES') is not None

    @property
    def subs(self):
        return self._subs

    @subs.setter
    def subs(self, value):
        non_strings =  \
            {k: v for k, v in value.items() if
             (not k == 'volume_host') and
             (not isinstance(v, six.string_types))}
        if len(non_strings) > 0:
            raise TypeError("subs can't contain non-string values:"
                            "\n%(non_strings)s" %
                            {'non_strings': non_strings})
        else:
            self._subs = value

    @classmethod
    def _get_sample_path(cls, name, dirname, suffix='', api_version=None):
        parts = [dirname]
        parts.append('samples')
        # Note: if _use_common_volume_api_samples is set to True
        # then common volume sample files present in 'volumes' directory
        # will be used. As of now it is being used for volume POST request
        # to avoid duplicate copy of volume req and resp sample files.
        # Example - VolumesSampleBase's _create_volume method.
        if cls._use_common_volume_api_samples:
            parts.append('volumes')
        else:
            parts.append(cls.sample_dir)
            if api_version:
                parts.append('v' + api_version)
        parts.append(name + ".json" + suffix)
        return os.path.join(*parts)

    @classmethod
    def _get_sample(cls, name, api_version=None):
        dirname = os.path.dirname(os.path.abspath(__file__))
        dirname = os.path.normpath(os.path.join(dirname,
                                                "../../../api-ref/source/v3"))
        return cls._get_sample_path(name, dirname, api_version=api_version)

    @classmethod
    def _get_template(cls, name, api_version=None):
        dirname = os.path.dirname(os.path.abspath(__file__))
        dirname = os.path.normpath(os.path.join(dirname,
                                   "./api_sample_tests"))
        return cls._get_sample_path(name, dirname, suffix='.tpl',
                                    api_version=api_version)

    def _read_template(self, name):
        template = self._get_template(name)
        with open(template) as inf:
            return inf.read().strip()

    def _write_template(self, name, data):
        with open(self._get_template(name), 'w') as outf:
            outf.write(data)

    def _write_sample(self, name, data):
        with open(self._get_sample(name), 'w') as outf:
            outf.write(data)

    def _compare_dict(self, expected, result, result_str, matched_value):
        if not isinstance(result, dict):
            raise NoMatch('%(result_str)s: %(result)s is not a dict.'
                          % {'result_str': result_str, 'result': result})
        ex_keys = sorted(expected.keys())
        res_keys = sorted(result.keys())
        if ex_keys != res_keys:
            ex_delta = []
            res_delta = []
            for key in ex_keys:
                if key not in res_keys:
                    ex_delta.append(key)
            for key in res_keys:
                if key not in ex_keys:
                    res_delta.append(key)
            raise NoMatch(
                'Dictionary key mismatch:\n'
                'Extra key(s) in template:\n%(ex_delta)s\n'
                'Extra key(s) in %(result_str)s:\n%(res_delta)s\n' %
                {'ex_delta': ex_delta, 'result_str': result_str,
                 'res_delta': res_delta})
        for key in ex_keys:
            res = self._compare_result(expected[key], result[key],
                                       result_str)
            matched_value = res or matched_value
        return matched_value

    def _compare_list(self, expected, result, result_str, matched_value):
        if not isinstance(result, list):
            raise NoMatch(
                '%(result_str)s: %(result)s is not a list.' %
                {'result_str': result_str, 'result': result})

        expected = expected[:]
        extra = []

        # if it's a list of 1, do the simple compare which gives a
        # better error message.
        if len(result) == len(expected) == 1:
            return self._compare_result(expected[0], result[0], result_str)

        # This is clever enough to need some explanation. What we
        # are doing here is looping the result list, and trying to
        # compare it to every item in the expected list. If there
        # is more than one, we're going to get fails. We ignore
        # those. But every time we match an expected we drop it,
        # and break to the next iteration. Every time we hit the
        # end of the iteration, we add our results into a bucket
        # of non matched.
        #
        # This results in poor error messages because we don't
        # really know why the elements failed to match each
        # other. A more complicated diff might be nice.
        for res_obj in result:
            for i, ex_obj in enumerate(expected):
                try:
                    matched_value = self._compare_result(ex_obj,
                                                         res_obj,
                                                         result_str)
                    del expected[i]
                    break
                except NoMatch:
                    pass
            else:
                extra.append(res_obj)

        error = []
        if expected:
            error.append('Extra list items in template:')
            error.extend([repr(o) for o in expected])

        if extra:
            error.append('Extra list items in %(result_str)s:' %
                         {'result_str': result_str})
            error.extend([repr(o) for o in extra])

        if error:
            raise NoMatch('\n'.join(error))
        return matched_value

    def _compare_template(self, expected, result, result_str, matched_value):
        # escape stuff for regex
        for char in '[]<>?':
            expected = expected.replace(char, '\\%s' % char)
        # special handling of subs that are not quoted. We are
        # expecting an int but we had to pass in a string
        # so the json would parse properly.
        if expected.startswith("%(int:"):
            result = str(result)
            expected = expected.replace('int:', '')

        expected = expected % self.subs
        expected = '^%s$' % expected
        match = re.match(expected, result)
        if not match:
            raise NoMatch(
                'Values do not match:\n'
                'Template: %(expected)s\n%(result_str)s: %(result)s' %
                {'expected': expected, 'result_str': result_str,
                 'result': result})
        try:
            matched_value = match.group()
        except IndexError:
            if match.groups():
                matched_value = match.groups()[0]
        return matched_value

    def _compare_result(self, expected, result, result_str):

        matched_value = None
        # None
        if expected is None:
            if result is None:
                pass
            else:
                raise NoMatch('%(result_str)s: Expected None, got %(result)s.'
                              % {'result_str': result_str, 'result': result})
        # dictionary
        elif isinstance(expected, dict):
            matched_value = self._compare_dict(
                expected, result, result_str, matched_value)
        # list
        elif isinstance(expected, list):
            matched_value = self._compare_list(
                expected, result, result_str, matched_value)
        # template string
        elif isinstance(expected, six.string_types) and '%' in expected:
            if expected[-1] == '%':
                if result != expected:
                    raise NoMatch(
                        'Values do not match:\n'
                        'Template: %(expected)s\n%(result_str)s: '
                        '%(result)s' % {'expected': expected,
                                        'result_str': result_str,
                                        'result': result})
            else:
                matched_value = self._compare_template(
                    expected, result, result_str, matched_value)
        # string
        elif isinstance(expected, six.string_types):

            # Ignore whitespace in this comparison
            expected = expected.strip()
            if isinstance(result, six.string_types):
                result = result.strip()

            if expected != result:
                raise NoMatch(
                    'Values do not match:\n'
                    'Template: %(expected)s\n%(result_str)s: '
                    '%(result)s' % {'expected': expected,
                                    'result_str': result_str,
                                    'result': result})
        # int
        elif isinstance(expected, (six.integer_types, float)):
            if expected != result:
                raise NoMatch(
                    'Values do not match:\n'
                    'Template: %(expected)s\n%(result_str)s: '
                    '%(result)s' % {'expected': expected,
                                    'result_str': result_str,
                                    'result': result})

        else:
            raise ValueError(
                'Unexpected type %(expected_type)s'
                % {'expected_type': type(expected)})

        return matched_value

    def generalize_subs(self, subs, vanilla_regexes):
        """Give the test a chance to modify subs after the server response

        was verified, and before the on-disk doc/samples file is checked.
        This may be needed by some tests to convert exact matches expected
        from the server into pattern matches to verify what is in the
        sample file.

        If there are no changes to be made, subs is returned unharmed.
        """
        return subs

    def _update_links(self, sample_data):
        """Process sample data and update version specific links."""
        # replace version urls
        url_re = self._get_host() + "/v3/" + PROJECT_ID
        new_url = self._get_host() + "/" + self.api_major_version
        if self._project_id:
            new_url += "/" + PROJECT_ID
        updated_data = re.sub(url_re, new_url, sample_data)

        # replace unversioned urls
        url_re = self._get_host() + "/" + PROJECT_ID
        new_url = self._get_host()
        if self._project_id:
            new_url += "/" + PROJECT_ID
        updated_data = re.sub(url_re, new_url, updated_data)
        return updated_data

    def _verify_response(self, name, subs, response, exp_code,
                         update_links=True):

        # Always also include the laundry list of base regular
        # expressions for possible key values in our templates. Test
        # specific patterns (the value of ``subs``) can override
        # these.
        regexes = self._get_regexes()
        regexes.update(subs)
        subs = regexes
        self.subs = subs
        message = response.text if response.status_code >= 400 else None
        self.assertEqual(exp_code, response.status_code, message)
        response_data = response.content
        response_data = pretty_data(response_data)
        if not os.path.exists(self._get_template(name)):
            self._write_template(name, response_data)
            template_data = response_data
        else:
            template_data = self._read_template(name)
        if (self.generate_samples and
                not os.path.exists(self._get_sample(name))):
            self._write_sample(name, response_data)
            sample_data = response_data
        else:
            with open(self._get_sample(name)) as sample:
                sample_data = sample.read()
                if update_links:
                    sample_data = self._update_links(sample_data)

        try:
            template_data = objectify(template_data)
            response_data = objectify(response_data)
            response_result = self._compare_result(template_data,
                                                   response_data, "Response")
        except NoMatch as e:
            raise NoMatch("\nFailed to match Template to Response: \n%s\n"
                          "Template: %s\n\n"
                          "Response: %s\n\n" %
                          (e,
                           pp.pformat(template_data),
                           pp.pformat(response_data)))

        try:
            # replace some of the subs with patterns for the
            # doc/samples check. Also let the test do some of its
            # own generalization, if necessary
            vanilla_regexes = self._get_regexes()
            subs['os-vol-host-attr:host'] = vanilla_regexes['host_name']
            subs['id'] = vanilla_regexes['id']
            subs['uuid'] = vanilla_regexes['uuid']
            subs['image_id'] = vanilla_regexes['uuid']
            subs = self.generalize_subs(subs, vanilla_regexes)
            self.subs = subs
            sample_data = objectify(sample_data)
            self._compare_result(template_data, sample_data, "Sample")
            return response_result
        except NoMatch as e:
            raise NoMatch("\nFailed to match Template to Sample: \n%s\n"
                          "Template: %s\n\n"
                          "Sample: %s\n\n"
                          "Hint: does your test need to override "
                          "ApiSampleTestBase.generalize_subs()?" %
                          (e,
                           pp.pformat(template_data),
                           pp.pformat(sample_data)))

    def _get_host(self):
        return 'http://localhost:8776'

    def _get_glance_host(self):
        return 'http://glance.openstack.example.com'

    def _get_regexes(self):
        text = r'(\\"|[^"])*'
        isotime_re = r'\d{4}-[0,1]\d-[0-3]\dT\d{2}:\d{2}:\d{2}Z'
        strtime_re = r'\d{4}-[0,1]\d-[0-3]\dT\d{2}:\d{2}:\d{2}\.\d{6}'
        extension_update = (
            r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}[+-]\d{2}:\d{2}')
        strtime_url_re = (r'\d{4}-[0,1]\d-[0-3]\d'
                          r'\+\d{2}\%3A\d{2}\%3A\d{2}\.\d{6}')

        return {
            'isotime': isotime_re,
            'strtime': strtime_re,
            'strtime_url': strtime_url_re,
            'strtime_or_none': r'None|%s' % strtime_re,
            'password': '[0-9a-zA-Z]{1,12}',
            'id': '(?P<id>[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}'
                  '-[0-9a-f]{4}-[0-9a-f]{12})',
            'uuid': '[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}'
                    '-[0-9a-f]{4}-[0-9a-f]{12}',
            'request_id': 'req-[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}'
                          '-[0-9a-f]{4}-[0-9a-f]{12}',
            'host': r'https?://[0-9]+(?:\.[0-9]+){3}:[0-9]+',
            'host_name': r'\w+',
            'glance_host': self._get_glance_host(),
            'os-vol-host-attr:host': self.volume.host,
            'text': text,
            'int': '[0-9]+',
            'user_id': text,
            'api_vers': self.api_major_version,
            'volume_endpoint': self._get_volume_endpoint(),
            'versioned_volume_endpoint': self._get_versioned_volume_endpoint(),
            'name': text,
            'description': text,
            'extension_update': extension_update,
            'auth_key': '[a-z0-9]{16}'
        }

    def _get_volume_endpoint(self):
        if self._project_id:
            return '%s/%s' % (self._get_host(), PROJECT_ID)
        else:
            return self._get_host()

    def _get_versioned_volume_endpoint(self):
        if self._project_id:
            return '%s/v3/%s' % (self._get_host(), PROJECT_ID)
        else:
            return self._get_host() + '/v3/'

    def _get_response(self, url, method, body=None,
                      headers=None):
        headers = headers or {}
        headers['Content-Type'] = 'application/json'
        headers['Accept'] = 'application/json'
        return self.api.api_request(url, body=body, method=method,
                                    headers=headers)

    def _do_options(self, url, headers=None):
        return self._get_response(url, 'OPTIONS', headers=headers)

    def _do_get(self, url, headers=None,
                return_json_body=False):
        response = self._get_response(url, 'GET', headers=headers)
        if return_json_body and hasattr(response, 'content'):
            return jsonutils.loads(response.content)
        return response

    def _do_post(self, url, name=None, subs=None, method='POST', headers=None):
        self.subs = {} if subs is None else subs
        body = None
        if name:
            body = self._read_template(name) % self.subs
            sample = self._get_sample(name)
            if self.generate_samples and not os.path.exists(sample):
                self._write_sample(name, body)
        return self._get_response(url, method, body, headers=headers)

    def _do_put(self, url, name=None, subs=None, headers=None):
        # name indicates that we have a body document. While the HTTP
        # spec implies that PUT is supposed to have one, we have some
        # APIs which don't.
        if name:
            return self._do_post(
                url, name, subs, method='PUT', headers=headers)
        else:
            return self._get_response(url, 'PUT', headers=headers)

    def _do_delete(self, url, headers=None):
        return self._get_response(url, 'DELETE', headers=headers)
